#!/bin/bash
#-----------------------------------------------------------------------------

if [ -z "${INCLUDED_HELPERSCRIPTS_SH}" ]; then
    readonly INCLUDED_HELPERSCRIPTS_SH=TRUE
#-----------------------------------------------------------------------------

export function log

export function get_os_product_version
export function trim
export function to_upper
export function to_lower
export function get_temp_file
export function test_variable
export function test_function
export function contains
export function test_tool
export function push_ifs
export function pop_ifs
export function arg_strip
export function format_uuid
export function get_uuids
export function get_uuid
export function get_file_size
export function escape_str
export function quote_item
export function quote_params_with_item_separator
export function create_find_params_from_pattern
export function search_protocol_handlers
export function search_default_handler

export function get_bundle_binary
export function read_bundle_info_key
export function write_bundle_info_key
export function write_info_plist_key
export function write_bundle_versions

export function urlencode_utf8_char
export function rawurlencode
export function rawxmlencode

export function update_config
#-----------------------------------------------------------------------------

export readonly FALSE=0
export readonly TRUE=1

# Supported log level values
export readonly LOG_ERROR=0
export readonly LOG_WARNING=1
export readonly LOG_INFO=2
export readonly LOG_DEBUG=3
export readonly LOG_TRACE=4

export readonly PATH_SEPARATOR=":"
export readonly EXCLUDE_PATTERN_CHAR="!"
export readonly APPLE_BUNDLEID_BASE="com.apple"
#-----------------------------------------------------------------------------

IFS_STACK_COUNTER=-1
declare -a IFS_STACK
#-----------------------------------------------------------------------------

function get_os_product_version()
{
    echo -e $(sw_vers -productVersion | awk -F. '{print $2}')
}
#-----------------------------------------------------------------------------

# Trim given chars from both ends of the input param
function trim()
{
    local CHARS="\t "
    if [[ $# -ge 2 ]]; then
        CHARS="$2"
    fi
    echo "$1" | sed -e "s/^[${CHARS}]*//g" -e "s/[${CHARS}]*$//g"
}
#-----------------------------------------------------------------------------

function to_upper()
{
    echo "$1" | tr '[:lower:]' '[:upper:]'
}
#-----------------------------------------------------------------------------

function to_lower()
{
    echo "$1" | tr '[:upper:]' '[:lower:]'
}
#-----------------------------------------------------------------------------

function test_variable()
{
    eval '[[ ${!'$1'[@]} ]]' && echo TRUE || echo FALSE
}
#-----------------------------------------------------------------------------

function test_function()
{
    #    type -t $function
    $(declare -fx $1 > /dev/null) && echo TRUE || echo FALSE
}
#-----------------------------------------------------------------------------

function contains()
{
    eval '[[ "$1" =~ "$2" ]]' && echo TRUE || echo FALSE
}
#-----------------------------------------------------------------------------

function test_tool()
{
# TODO: find a proper solution that works for each tool, none of -v or -h behave the same way :/ f.e. atos always returns none 0, even if -h used
#   local TOOL="${1}"
#
#    if [ -f "${TOOL}" ]; then
#        "${TOOL}" --version 1>/dev/null 2>&1 && echo TRUE || echo FALSE
#    else
#        echo FALSE
#    fi
    echo TRUE
}
#-----------------------------------------------------------------------------

function get_temp_file()
{
    local FILE_BASE_NAME="${1}"
    local TMP_FILE=`mktemp -t "${FILE_BASE_NAME}"`; # Let the shell create a temporary file
    
    echo "${TMP_FILE}"
}
#-----------------------------------------------------------------------------

function push_ifs()
{
    ((IFS_STACK_COUNTER++))
    IFS_STACK[${IFS_STACK_COUNTER}]=${IFS}
    #log LOG_TRACE IFS_STACK_COUNTER=$IFS_STACK_COUNTER IFS_STACK[${IFS_STACK_COUNTER}]=${IFS_STACK[${IFS_STACK_COUNTER}]}

    if [[ "${1}" == "" ]]; then
        #unset IFS  # Not the same!
        IFS=""      # This is treated as null
    else
        IFS=${1}
    fi
    return 0
}
#-----------------------------------------------------------------------------

function pop_ifs()
{
    #log LOG_TRACE IFS_STACK_COUNTER=$IFS_STACK_COUNTER IFS_STACK[${IFS_STACK_COUNTER}]=${IFS_STACK[${IFS_STACK_COUNTER}]}
    IFS=${IFS_STACK[${IFS_STACK_COUNTER}]}
    ((IFS_STACK_COUNTER--))
    return 0    # Should use here, ((IFS_STACK_COUNTER--)) could cause returning with f.e. with a negative number
}
#-----------------------------------------------------------------------------

function arg_strip()
{
    tr -d '\r\n' | sed -e 's/^ *//g' -e 's/ *$//g'
}
#-----------------------------------------------------------------------------

function format_uuid()
{
    local UUID=$1
    echo ${UUID} | tr -d '-'                                                    # older dumps has no - separators in the UUID and much more easier to unformat, so removing any separator
}
#-----------------------------------------------------------------------------

#
# Could produce multiple uuids separated by \n for unibin targets
#
function get_uuids()
{
    local BINARY="${1}"
    dwarfdump --uuid "${BINARY}" | awk '{ printf("%s\n", $2); }'                # it should be well enough here, but who knows why does not work on 10.7.5  grep -i -o [0-9a-fA-F]*\-[0-9a-fA-F]*\-[0-9a-fA-F]*\-[0-9a-fA-F]*\-[0-9a-fA-F]*
}
#-----------------------------------------------------------------------------

#
# Could produce only one uuid per architecture
#
function get_uuid()
{
    local BINARY="${1}"
    local ARCH="${2}"
    
    if [ -n "${BINARY}" ] && [ -n "${ARCH}" ]; then
	    dwarfdump --uuid "${BINARY}" | grep -i ${ARCH}| awk '{ printf("%s\n", $2); exit; }' 
	fi
}
#-----------------------------------------------------------------------------

function get_file_size()
{
    local SIZE=0

    if [[ -f "${1}" ]]; then
        local RESULT=$(/bin/ls -al "${1}" | awk '{ print $5 }' 2>&1)
        if [[ -n "${RESULT}" && $(contains "${RESULT}" "ls:") == FALSE ]]; then
            SIZE=${RESULT}
        fi
    fi
    echo ${SIZE}
}
#-----------------------------------------------------------------------------

function escape_str()
{
    echo -en "${1}" | sed \
        -e 's/\//\\\//g' \
        -e 's/\&/\\\&/g' \
        -e 's/\$/\\\$/g' \
        -e 's/\"/\\\"/g'
}
#-----------------------------------------------------------------------------

function quote_item()
{
    local QUOTE_CHAR=${1}
    local WHAT=${2}
    local WHAT_LEN=${#WHAT}

    #log LOG_TRACE "quote_item called with QUOTE_CHAR=\"${QUOTE_CHAR}\", WHAT=\"${WHAT}\", WHAT_LEN=${WHAT_LEN}" > /dev/null # Log into log file only!

    if [[ "${WHAT}" == "" || "${WHAT}" == "${QUOTE_CHAR}" ]]; then
        WHAT="${QUOTE_CHAR}${QUOTE_CHAR}"
    else
        #log LOG_TRACE "WHAT:0:1=\"${WHAT:0:1}\", WHAT:LEN-1=${WHAT:${WHAT_LEN}-1:1}" > /dev/null # Log into log file only!
        if [[ "${WHAT:${WHAT_LEN}-1:1}" != "${QUOTE_CHAR}" ]]; then
            WHAT="${WHAT}${QUOTE_CHAR}"
        fi
        if [[ "${WHAT:0:1}" != "${QUOTE_CHAR}" ]]; then
            WHAT="${QUOTE_CHAR}${WHAT}"
        fi
    fi
    #log LOG_TRACE "quote_item result is:${WHAT}" > /dev/null # Log into log file only!
    echo "${WHAT}"
}
#-----------------------------------------------------------------------------

function quote_params_with_item_separator()
{
    local INPUT_SEPARATOR="${1}"
    local OUTPUT_SEPARATOR="${2}"
    shift 2

    push_ifs "${INPUT_SEPARATOR}"
    local ELEMENTS=("$@")
    local ELEMENT
    local RESULT

    log LOG_TRACE "quote_params_with_item_separator called with INPUT_SEPARATOR=\"${INPUT_SEPARATOR}\", OUTPUT_SEPARATOR=\"${OUTPUT_SEPARATOR}\", ${#ELEMENTS[@]} element(s) in array ELEMENTS=${ELEMENTS[@]}" > /dev/null # Log into log file only!

    for ELEMENT in ${ELEMENTS[@]}; do
        #log LOG_TRACE "quoting element:${ELEMENT}" > /dev/null # Log into log file only!
        if [[ -z "${RESULT}" ]]; then
            RESULT=$(quote_item "\"" "${ELEMENT}")
        else
            RESULT=${RESULT}${OUTPUT_SEPARATOR}$(quote_item "\"" "${ELEMENT}")
        fi
    done

    pop_ifs
    echo "${RESULT}"
}
#-----------------------------------------------------------------------------

function create_find_params_from_pattern
{
    local OUTPUT=$1
    local PATTERNS=$2
    local NEXT_PATTERN
    local EXCLUDE_PATTERN

    push_ifs $"${PATH_SEPARATOR}"
	set -f
	
    for NEXT_PATTERN in ${PATTERNS}; do
        if [[ "${NEXT_PATTERN:0:1}" == "${EXCLUDE_PATTERN_CHAR}" ]]; then
            if [ -n "${EXCLUDE_PATTERN}" ]; then
                EXCLUDE_PATTERN=${EXCLUDE_PATTERN}" -and "
            fi
            EXCLUDE_PATTERN="${EXCLUDE_PATTERN}-name ${NEXT_PATTERN:1}"
        else
            if [ -n "${OUTPUT}" ]; then
                OUTPUT=${OUTPUT}" -or "
            fi
            OUTPUT="${OUTPUT}-name ${NEXT_PATTERN}"
        fi
    done
    if [ -n "${OUTPUT}" ] && [ -n "${EXCLUDE_PATTERN}" ]; then
        OUTPUT="! ( "${EXCLUDE_PATTERN}" ) -and ( "${OUTPUT}" )"
    fi

	set +f
    pop_ifs

    echo "${OUTPUT}"
}
#-----------------------------------------------------------------------------

function search_protocol_handlers()
{
    local PROTOCOL=$1
    local SEPARATOR=$( [ "$2" == "" ] && echo "\n" || echo "$2" )
    local REGREADER="/System/Library/Frameworks/CoreServices.framework/Versions/A/Frameworks/LaunchServices.framework/Versions/A/Support/lsregister"

    if [ "${PROTOCOL}" != "" ]; then
        "${REGREADER}" -dump | awk -v protocol="${PROTOCOL}" -v separator="${SEPARATOR}" '
            function reset() {
                inBundle = 0;
                bundleID = "";
                bundleIdentifier = "";
                bundleName = "";
                bundleVersion= "";
                bindings="";
            }
    
            function endOfBundle() {
                #foundProtocolHandler();
                #printf("end of bundle with id: %s\n", bundleIdentifier);

                reset();
            }
    
            function foundProtocolHandler() {
                #
                # TODO: Add version later, now chars like () kills writing into defaults
                #
                #printf("\"%s\"%c\"%s --%s==\"%c", bundleIdentifier, separator, bundleName, bundleVersion, separator);
                printf("\"%s\"%c\"%s\"%c", bundleIdentifier, separator, bundleName, separator);
            }

            BEGIN {
                inBundle = 0;
                bundleID = "";
                bundleIdentifier = "";
                bundleName = "";
                bundleVersion= "";
                bindings="";
            }
            /^bundle[\t ]*id:/{
                reset();
                inBundle = 1;
                bundleID = $3;
                #printf("found bundle with id: %s\n", bundleID);
            }
            /^[\t ]*name:/{
                if (bundleName == "") # further sub items could have also name: field, we need the first taht belongs the bundle desription
                    bundleName = substr($0, index($0, $2), length($0));
            }
            /^[\t ]*version:/{
                if (bundleVersion == "") # further sub items could have also name: field, we need the first taht belongs the bundle desription
                    bundleVersion = $2;
            }
            /^[\t ]*identifier:/{
                bundleIdentifier = $2;
            }
            /^[\t ]*bindings:/{
                bindings = $2;
                if (bindings == protocol)
                    foundProtocolHandler();
            }
            /^-*$/{
                if (inBundle != 0) {
                    endOfBundle();
                }
            }
            {
            }
            END {
            }
        '
    fi
}
#-----------------------------------------------------------------------------

# NOTE: Lazy implementation, does not care the container
#           LSHandlers =     (
# line at all
function search_default_handler()
{
    local PROTOCOL=$1
    local DEFAULT_IF_NOT_FOUND=$2
    local SOURCE=$( [[ ${OS_PRODUCT_VERSION} -gt 9 ]] && echo -n "${APPLE_BUNDLEID_BASE}.LaunchServices/${APPLE_BUNDLEID_BASE}.LaunchServices.secure" || echo -n "${APPLE_BUNDLEID_BASE}.LaunchServices" )
    # TODO: Use the one bellow instead here too
    #local REGREADER="/System/Library/Frameworks/CoreServices.framework/Versions/A/Frameworks/LaunchServices.framework/Versions/A/Support/lsregister"

    if [ "${PROTOCOL}" != "" ]; then
        defaults read ${SOURCE} | awk -v protocol="${PROTOCOL}" -v default_to="${DEFAULT_IF_NOT_FOUND}" '
            function endOfElement() {
                inElement = 0;
                bundleIdentifier = "";
                protocol_matches = 0;
            }
    
            BEGIN {
                endOfElement();
            }
            /^[\t ]*\{/{
                inElement = 1;
                protocol_matches = 0;
                bundleIdentifier = "";
            }
            /^[\t ]*(LSHandlerContentType|LSHandlerURLScheme)/{
                if (protocol == substr($3, 0, length($3) - 1)) {
                    protocol_matches = 1;
                    if (bundleIdentifier != "")  {   # already read the bundle id
                        exit;
                    }
                }
            }
            /^[\t ]*LSHandlerRoleAll/{
                bundleIdentifier = $3;
                #printf("protocol_matches: %d\n", protocol_matches );
                if (protocol_matches != 0) {   # already read the protocoll too and matches
                    exit;
                }
            }
            /^[\t ]*\}/{
                endOfElement();
            }
            {
            }
            END {
                if (bundleIdentifier == "") {
                    bundleIdentifier = default_to;
                }
                else {
                    bundleIdentifier = substr(bundleIdentifier, 0, length(bundleIdentifier) - 1);
                }
                printf("%s", bundleIdentifier);
            }
        '
    fi
}
#-----------------------------------------------------------------------------

# Read a bundle info plist item data by key
function read_bundle_info_key()
{
    local BUNDLE=$1
    local KEY=$2
    local INFO_PATH="$BUNDLE/Contents/Info"

    echo $(defaults read "$INFO_PATH" $KEY) || error_exit ERR_CALL "Couldn't read key ($KEY) of bundle ($INFO_PATH)"
    return 0
}
#-----------------------------------------------------------------------------

function get_bundle_binary()
{
    local BUNDLE_PATH=$1
    #local BASE_NAME=$(basename "$BUNDLE_PATH")

    #echo "${BUNDLE_PATH}/Contents/MacOS/$BASE_NAME"          # TODO: parse the application binary from the bundle info.plist based on CFBundlePackageType, this is good only for APPL type
    # can be ${BUNDLE_PATH}/Contents/Resources/DWARF/$BASE_NAME  if CFBundlePackageType is dSYM
    
    read_bundle_info_key "${BUNDLE_PATH}" "CFBundleExecutable"
}
#-----------------------------------------------------------------------------

function get_bundle_binary_path()
{
    local BUNDLE_PATH=$1

    #local BASE_NAME=$(basename "$BUNDLE_PATH")
    #echo "${BUNDLE_PATH}/Contents/MacOS/$BASE_NAME"          # TODO: parse the application binary from the bundle info.plist based on CFBundlePackageType, this is good only for APPL type
    # can be ${BUNDLE_PATH}/Contents/Resources/DWARF/$BASE_NAME  if CFBundlePackageType is dSYM
    
    local BINARY_NAME=$(read_bundle_info_key "${BUNDLE_PATH}" "CFBundleExecutable")
    echo "${BUNDLE_PATH}/Contents/MacOS/${BINARY_NAME}"
    return 0
}
#-----------------------------------------------------------------------------

# Write a bundle info plist item data by key
function write_bundle_info_key()
{
    local BUNDLE=$1
    local KEY=$2
    local DATA=$3
    local INFO_PATH="$BUNDLE/Contents/Info"

    defaults write "$INFO_PATH" "$KEY" -string "$DATA" || error_exit ERR_CALL "Couldn't write key ($KEY) of bundle ($INFO_PATH)"
    return 0
}
#-----------------------------------------------------------------------------

# Write a info plist item data by key
function write_info_plist_key()
{
    local INFO_PATH=$1
    local KEY=$2
    local DATA=$3
    local PLIST_NAME=$(basename "$INFO_PATH")
    local PLIST_POS=$(expr $PLIST_NAME : ".*\.plist$")

    if [ $PLIST_POS ]; then
        PLIST_NAME=${PLIST_NAME%.plist}
        INFO_PATH=$(dirname "$INFO_PATH")/$PLIST_NAME
    fi
    echo "Writing to domain/plist $PLIST_NAME at path $INFO_PATH into key $KEY data $DATA"

    defaults write "$INFO_PATH" "$KEY" -string "$DATA" || error_exit ERR_CALL "Couldn't write key ($KEY) of bundle ($INFO_PATH)"
    return 0
}
#-----------------------------------------------------------------------------

# Write version strings back in the given bundle
function write_bundle_versions()
{
    local BUNDLE="$1"
    local VERSION="$2"
    local INFO_PATH="$BUNDLE/Contents/Info"

    write_bundle_info_key "$BUNDLE" CFBundleShortVersionString "${VERSION}"
    #write_bundle_info_key "$BUNDLE" CFBundleVersion "${VERSIONS[$VER_KEY_CF_BUNDLE_VERSION]}"
    #write_bundle_info_key "$BUNDLE" CFBundleGetInfoString "${VERSIONS[$VER_KEY_CF_BUNDLE_GET_INFO_STRING]}"

    plutil -convert xml1 "$INFO_PATH.plist"
    chmod 0444 "$INFO_PATH.plist"
    return 0
}
#-----------------------------------------------------------------------------

function set_icon_from_res()
{
    local ITEM_TO_ICONIZE="$1"
    local ICON_RES_INPUT="$2"
    local ICONIZED_ITEM="$ITEM_TO_ICONIZE"

    if [ -d "$ITEM_TO_ICONIZE" ]; then
        ICONIZED_ITEM="$ITEM_TO_ICONIZE/"$'Icon\r'
    fi

    # Append a resource to the file or folder/Icon\r you want to icon-ize.
    Rez -append "$ICON_RES_INPUT" -o "$ICONIZED_ITEM"

    # Use the resource to set the icon.
    SetFile -a C "$ITEM_TO_ICONIZE"

    # Hide the Icon\r file from Finder if it's a dir
    if [ -d "$ITEM_TO_ICONIZE" ]; then
        SetFile -a V "$ICONIZED_ITEM"
    fi
}
#-----------------------------------------------------------------------------

function set_icon()
{
    local ITEM="$1"
    local INPUT_ICON="$2"
    local ICON=$(basename "$INPUT_ICON")
    local ICON_RES="$PROJ_TEMP_DIR/${ICON}.rsrc"

    # Take an image and make the image its own icon.
    local RESULT=$(sips -i "$INPUT_ICON" --out "$PROJ_TEMP_DIR/$ICON" 2>&1)
    #echo "$RESULT"

    # Extract the icon to its own resource file.
    DeRez -only icns "$PROJ_TEMP_DIR/$ICON" > "$ICON_RES"

    set_icon_from_res "$ITEM" "$ICON_RES"

    #Cleanup
    rm "$ICON_RES"
}
#-----------------------------------------------------------------------------

# Replaces given patterns with given string in a given file
# BUG: Calling this with 2 pattern params only breaks the for loop silently, could not yet figured out why :/
function expand_variables_in_file()
{
    local SOURCE="$1"
    local TARGET="$2"
    shift 2
    local PATTERNS=("${@}")
    local PATTERN_NUM=$#
    local PNDX=0
    local TARGET_TMP=""

    #echo "Patter num="$PATTERN_NUM
    #echo $SOURCE
    #echo $TARGET

    # TODO: Optimize this, do it in one pass
    for (( PNDX=0; PNDX<$PATTERN_NUM; PNDX=PNDX+2 )); do
        #echo entering for pndx=$PNDX
        local INP="${PATTERNS[$PNDX]}"
        local OUT="${PATTERNS[$PNDX+1]}"
        local CLEAR=0
        #echo "INP=$INP"
        #echo "OUT=$OUT"
        TARGET_TMP=$TARGET"_"$PNDX

        if [ "${OUT}" == "#{CLEAR_LINE}" ]; then
            CLEAR=1
        fi
        awk -v clear=${CLEAR} '{ orig_line=$0; gsub(/'"$INP"'/, "'"$OUT"'"); if (orig_line == $0 || !clear) print $0; }' "$SOURCE" > "$TARGET_TMP"
        #echo awked
        SOURCE=$TARGET_TMP
        #echo src=tmp done
    done
    #echo tmp->target
    mv -f "$TARGET_TMP" "$TARGET"
    #echo tmp->target done
    rm "$TARGET"_* 2>/dev/null
}
#-----------------------------------------------------------------------------

# Get system date time in a given format (see 'man date/strftime' for formatting strings)
function get_date_time_formatted()
{
    date -j "+$1"
}
#-----------------------------------------------------------------------------

# Get system date time in a given format (see 'man date/strftime' for formatting strings)
function get_date_time_formatted()
{
    local FMT="+$1"
    
    if [ "${FMT}" == "+" ]; then
        FMT="+%Y-%m-%d_%H:%M:%S"
    fi
    
    date -j "${FMT}"
}
#-----------------------------------------------------------------------------

function urlencode_utf8_char()
{
    local OUT=$(printf "%%%s" $(echo -en "${1}" | xxd -p -c1 -u))
    #log LOG_TRACE "urlencode_utf8_char result is CHAR='${CHAR}', OUT='${OUT}'" > /dev/null
    # NOTE:  Unable to solve the problem that OUT formatting leads to \n conversion of each % after the first occurence
    #        IFS has affect on it, but could not reset to a proper value
    #        Be sure that IFS has the defaults when calling here!!!
    #        A workaround can be transforming the linefeeds to % afterwards
    echo -en "${OUT}" #| tr '\n' '%%'
}
#-----------------------------------------------------------------------------

function rawurlencode()
{
    local STRING="${1}"
    local ADD_LINE_END="${2}"
    local STRLEN=${#STRING}
    local ENCODED=""

    if [[ ${STRLEN} > 0 ]]; then

        for (( POS=0 ; POS<STRLEN ; POS++ )); do
            local CHAR="${STRING:${POS}:1}"
            local OUT

            case "${CHAR}" in
                [-_.~a-zA-Z0-9] )
                    OUT="${CHAR}"
                ;;

                * )
                    OUT=$(urlencode_utf8_char "${CHAR}")    # do not forget to quote, because of chars like space f.e.
                    #log LOG_TRACE "Returning result of urlencode_utf8_char: POS=${POS}, CHAR='${CHAR}', OUT='${OUT}'" > /dev/null
                ;;
            esac
            ENCODED+="${OUT}"
        done
        if [[ ${ADD_LINE_END} == TRUE ]]; then
            ENCODED+="%0a"
        fi
        echo -en "${ENCODED}"
    fi
}
#-----------------------------------------------------------------------------

function rawxmlencode()
{
    local STRING="${1}"
    local ADD_LINE_END="${2}"
    local STRLEN=${#STRING}
    local ENCODED=""

    if [[ ${STRLEN} > 0 ]]; then

        for (( POS=0 ; POS<STRLEN ; POS++ )); do
            local CHAR="${STRING:${POS}:1}"
            local OUT

            case "${CHAR}" in
                \< )
                    OUT="&#60;"
                ;;

                \> )
                    OUT="&#62;"
                ;;

                \" )
                    OUT="&#34;"
                ;;

                \& )
                    OUT="&#38;"
                ;;

                \' )
                    OUT="&#39;"
                ;;

                * )
                    OUT="${CHAR}"
                ;;
            esac
            ENCODED+="${OUT}"
        done
        if [[ ${ADD_LINE_END} == TRUE ]]; then
            ENCODED+="%0a"
        fi
        echo -en "${ENCODED}"
    fi
}
#-----------------------------------------------------------------------------

#
# PLIST_VAR_PREFIX and VAR_NAMES should be preset at call time :/
#
function update_config()
{
    CONF_FILE="${1}"
    CONF_NEW_FILE="${2}"
    DEFAULTS_DOMAIN=
    if [ -n "${3}" ]; then
        DEFAULTS_DOMAIN="${3}"
    fi

    CONF_FILE_NAME_BACKUP=$(basename "${CONF_FILE}")
    CONF_BACKUP_DIR=$(dirname "${CONF_FILE}")/.Backup
    CONF_FILE_BACKUP="${CONF_BACKUP_DIR}/${CONF_FILE_NAME_BACKUP}"
    CONF_FILE_TMP="${CONF_FILE}.tmp"
    CONF_DIR=$(dirname "${CONF_FILE}")

    #echo update_config called with: $@
    #echo CONF_FILE=${CONF_FILE}
    #echo CONF_FILE=${CONF_FILE}
    #echo CONF_NEW_FILE=${CONF_NEW_FILE}
    #echo CONF_FILE_BACKUP=${CONF_FILE_BACKUP}
    #echo CONF_FILE_TMP=${CONF_FILE_TMP}
    #echo CONF_DIR=${CONF_DIR}
    #echo CONF_BACKUP_DIR=${CONF_BACKUP_DIR}
    #echo DEFAULTS_DOMAIN=${DEFAULTS_DOMAIN}

    if ! [ -d "${CONF_DIR}" ]; then
        mkdir "${CONF_DIR}"
    fi
    if ! [ -d "${CONF_BACKUP_DIR}" ]; then
        mkdir "${CONF_BACKUP_DIR}"
    fi

    echo "Updating ${CONF_FILE}"
    # Try to backup user defined config if any
    if [ -f "${CONF_FILE}" ]; then
        echo "Backing up config..."
        mv -fv "${CONF_FILE}" "${CONF_FILE_BACKUP}"
    else
        echo "No old config file found, using default config as backup."
        cp -fv "${CONF_NEW_FILE}" "${CONF_FILE_BACKUP}"
    fi
    cp -fv "${CONF_NEW_FILE}" "${CONF_FILE}"

    echo "Updating config based on backup…"

    declare PATTERNS
    #echo "# ---DEBUG--- VAR_NAMES=${VAR_NAMES[@]}"
    for VAR in "${VAR_NAMES[@]}"; do
        local THIS_PATTERN="^[\t ]*(export[\t ]+)*(readonly[\t ]+)*${VAR}[\t ]*="

        if [ -z "${PATTERNS}" ]; then
            PATTERNS="${THIS_PATTERN}"
        else
            PATTERNS="${PATTERNS}\n${THIS_PATTERN}"
        fi
    done
    #echo -e "# ---DEBUG--- PATTERNS=\n${PATTERNS}"

    awk -v default_file="${CONF_NEW_FILE}"  -v original_file="${CONF_FILE_BACKUP}"  -v patterns="${PATTERNS}"  -v defaults_domain="${DEFAULTS_DOMAIN}" -v plist_var_prefix="${PLIST_VAR_PREFIX}" '
        function ltrim(s) { sub(/^[ \t\r\n]+/, "", s); return s; }
        function rtrim(s) { sub(/[ \t\r\n]+$/, "", s); return s; }
        function trim(s)  { return rtrim(ltrim(s)); }
        function escape(s)  { return  gsub("$", "\\$", s); }
        function read_from_file(file, what) {
            found_line="";
            while ((getline line < file) > 0) {
                if (match(line, what) != 0) {
                    found_line = line;
                    break;
                }
            }
            close(file);
            return found_line;
        }
        function get_line_value(line) {
            value="";
            if (length(line) != 0) {
                element_num = split(line, line_elements, "=");

                if (element_num > 0) {
                    value = trim(line_elements[element_num]);
                }
            }
            return value;
        }
        function get_line_variable(line) {
            variable="";
            if (length(line) != 0) {
                element_num = split(line, line_elements, "=");

                if (element_num > 0) {
                    #for(i = element_num; i >= 1; i--)
                    #   printf("# ---DEBUG--- %d of %d is \"%s\"\n", i, element_num, line_elements[i]);
                    expr_element_num = split(line_elements[element_num - 1], expr_elements, " ");

                    if (expr_element_num > 0) {
                        variable = expr_elements[expr_element_num];
                    }
                }
            }
            return variable;
        }
        BEGIN {
            pattern_num = split(patterns, pattern_array, "\n");
            #printf("# ---DEBUG--- pattern num is %d of \n\"%s\"\n", pattern_num, patterns);
        }
        {
            for (i = 1; i <= pattern_num; i++) {
                #printf("# ---DEBUG--- %d of %d is \"%s\"\n", i, pattern_num, pattern_array[i]);
                if (match($0, pattern_array[i]) != 0) {
                    original_line = read_from_file(original_file, pattern_array[i]);
                    default_line = read_from_file(default_file, pattern_array[i]);
                    default_value = get_line_value(default_line);

                    if (length(original_line) != 0) {
                        value = get_line_value(original_line);
                        variable = get_line_variable(original_line);
                    }
                    else {
                        value = default_value;
                        variable = get_line_variable(default_line);
                    }
                    #printf("# ---DEBUG--- found variable=%s, value=%s\n", variable, value);
                    if (length(variable) != 0) {
                        if (defaults_domain) {  # in the sandboxed version  defaults_domain must not be empty and should use  NSUserDefaults to store/read the real values
                            value = trim(value);
                            # TODO: use apostrophe to signal what it really should have to do, no evaluation at all
                            #       for example <apostrophe>Archive.$(date -j "+%Y-%m-%d_%H-%M-%S")<apostrophe> should go directly into defaults
                            #       but defaults syntax still should be figured out to eat f.e. that string
                            #
                            #if (substr(value, 0, 1) == "\x27") {
                            #    value = escape(value);
                            #    quoter = "";
                            #}
                            #else
                                quoter = "\"";

                            default_value_create_str = "|| (defaults write " defaults_domain " " plist_var_prefix variable " " quoter default_value quoter "; echo " quoter default_value quoter ")";

                            if (match(value, "(defaults[ \t]*read)+") == 0) {   # not already using defaults, then first write into defaults the current value
                                cmd = "defaults write " defaults_domain " " plist_var_prefix variable " " quoter value quoter;
                                printf("# ---DEBUG--- defaults cmd: \"%s\"\n", cmd);
                                system(cmd);
                            }
                            # Make sure the new line contains the correct defaults solution, export var = read_from_defaults if not found write_default_value_into_defaults and assign to var
                            original_line = "export readonly " variable "=$(defaults read " defaults_domain " " plist_var_prefix variable " " default_value_create_str ")";
                        }
                        else
                            original_line = "export readonly " variable "=" value;
                    }
                    printf("%s\n", original_line);
                    next;
                }
            }
            printf("%s\n", $0);
        }
    ' "${CONF_FILE}" > "${CONF_FILE_TMP}"

    mv -fv "${CONF_FILE_TMP}" "${CONF_FILE}"
    echo "Finished updating ${CONF_FILE}"
}
#-----------------------------------------------------------------------------

fi  #  if [ -z "${INCLUDED_HELPERSCRIPTS_SH}" ]; then
#-----------------------------------------------------------------------------
